package org.sc3d.apt.sss.v3;

/** A subclass of Token that represents a string of zero or more Tokens enclosed in brackets. The 'type' field of instances of Bracket will be 'TYPE_BRACKET', 'TYPE_SQUARE' or 'TYPE_BRACE'. */
public class Bracket extends Token {
  /** Constructs a Bracket.
   * @param opening the Token which represents the open bracket.
   * @param tokens the Tokens inside the brackets. This array is not copied away, and must therefore not be modified hereafter.
   * @param closing the Token which represents the close bracket.
   */
  public Bracket(
    Sentence sentence,
    Token opening,
    Token[] tokens,
    Token closing
  ) {
    super(
      calculateType(opening),
      sentence,
      opening.start,
      closing.start+closing.length-opening.start
    );
    this.opening = opening; this.closing = closing;
    this.length = tokens.length; this.tokens = tokens;
  }
  
  /* New API. */
  
  /** One of the bracket Tokens passed to the constructor. */
  public final Token opening, closing;
  
  /** The number of Tokens nested inside this Bracket. */
  public final int length;
  
  /** Returns the Token with index 'index' inside this Bracket. */
  public Token get(int index) { return this.tokens[index]; }
  
  /** Returns the Tokens inside this Bracket in a fresh array. */
  public Token[] toArray() { return (Token[])this.tokens.clone(); }
  
  /** Returns 'true' if 'opening' is an open round bracket and 'closing' is a close round bracket, or if 'opening' is an open square bracket and 'closing' is a close square bracket, or if 'opening' is an open brace and 'closing' is a close brace. Returns 'false' otherwise. */
  public static boolean matches(Token opening, Token closing) {
    if (opening.length!=1 || closing.length!=1) return false;
    final char c1 = opening.sentence.get(opening.start);
    final char c2 = closing.sentence.get(closing.start);
    return (c1=='(' && c2==')') || (c1=='[' && c2==']') || (c1=='{' && c2=='}');
  }
  
  /* Private. */
  
  /** Returns 'TYPE_BRACKET', 'TYPE_SQUARE' or 'TYPE_BRACE'. */
  private static int calculateType(Token opening) {
    final char c = opening.sentence.get(opening.start);
    if (c=='(') return TYPE_ROUND;
    if (c=='[') return TYPE_SQUARE;
    if (c=='{') return TYPE_BRACE;
    throw new IllegalArgumentException("Not an open bracket character");
  }
  
  private final Token[] tokens;
}
