package org.sc3d.apt.sss.v3;

/** Represents an SSS grammar. I will explain the data structure used by example. Here is an SSS grammar for ordinary four-operation arithmetic expressions (it is the one used by the Calculator example):<pre>
 * sign ::= {Minus {"-"}}
 * atom ::= {Number {sign* NUMBER} Bracket {sign* BRACKET(sum)}}
 * multiplicand ::= {Multiply {"*" atom} Divide {"/" atom}}
 * product ::= {Product {atom multiplicand*}}
 * summand ::= {Add {"+" product} Subtract {"-" product}}
 * sum ::= {Sum {product summand*}}
 * ROOT sum
 * </pre>I'm sorry it's quite complicated, but I want to illustrate as many of the cases as possible, and I want to use an example that is familiar. Here is an example string of this language, followed by the same string annotated with square brackets to show how it fits the grammar:<pre>
 *     - 5  *  - (     8     +    1      ) /   3    +    7    -    2
 * [[[[-]5][*[[-]([[[[]8][]][+[[[]1][]]]])]/[[]3]]][+[[[]7][]]-[[[]2][]]]]
 * </pre>I hope that's enough for you to understand the example grammar. Note that all the operator binding rules are as you'd expect, and that the "-" characters have been parsed in different ways (two of them are negations and the third is a subtraction).
 * <p>Here is the code needed to construct a Grammar to represent the above grammar and assign it to a static final field:<pre>
 * static final Grammar SUM {
 * static {
 * // First make all the Terminals we need.
 * Terminal plus = new Keyword("+");
 * Terminal minus = new Keyword("-");
 * Terminal multiply = new Keyword("*");
 * Terminal divide = new Keyword("/");
 * Terminal number = new Terminal(Token.TYPE_NUMBER);
 * Terminal bracket = new Terminal(Token.TYPE_BRACKET) {
 *   public Grammar getGrammarOfContents() { return SUM; }
 * };
 * // Construct 'signStar'.
 * Production pMinus = new Production("Minus", new Grammar[] {minus});
 * NonTerminal signStar = new NonTerminal(new Production[] {pMinus}, true, true);
 * // Construct 'atom'.
 * Production pNumber = new Production("Number", new Grammar[] {signStar, number});
 * Production pBracket = new Production("Bracket", new Grammar[] {signStar, bracket});
 * NonTerminal atom = new NonTerminal(new Production[] {pNumber, pBracket}, false, false);
 * // Construct 'multiplicandStar'.
 * Production pMultiply = new Production("Multiply", new Grammar[] {multiply, atom});
 * Production pDivide = new Production("Divide", new Grammar[] {divide, atom});
 * NonTerminal multiplicandStar = new NonTerminal(new Production[] {pMultiply, pDivide}, true, true);
 * // Construct 'product'.
 * Production pProduct = new Production("Product", new Grammar[] {atom, multiplicandStar});
 * NonTerminal product = new NonTerminal(new Production[] {pProduct}, false, false);
 * // Construct 'summandStar'.
 * Production pAdd = new Production("Add", new Grammar[] {plus, product});
 * Production pSubtract = new Production("Subtract", new Grammar[] {minus, product});
 * NonTerminal summandStar = new NonTerminal(new Production[] {pAdd, pSubtract}, true, true);
 * // Construct 'sum'.
 * Production pSum = new Production("Sum", new Grammar[] {product, summandStar});
 * NonTerminal sum = new NonTerminal(new Production[] {pSum}, false, false);
 * // The root is 'sum'.
 * SUM = sum;
 * }
 * </pre>As you see it is really quite a lot of work to construct a Grammar by hand. It is much easier to write an SSS grammar specification and use the GrammarParser.
 * <p>The example Grammar is provided as a public static final field of the Calculator class, and can be used as a test case in debugging.
 */
public abstract class Grammar {
  /** Don't want any instances except of the inner classes. */
  private Grammar(boolean isTerminal) { this.isTerminal = isTerminal; }
  
  /* New API. */
  
  /** 'true' if this Grammar is an instance of Terminal, otherwise 'false'. */
  public final boolean isTerminal;
  
  /** Returns a String representation of this Grammar. */
  public abstract String toString();
  
  ////////////////////////////////////////////////////////////////////////////
  
  /** The subclass of Grammar with 'isTerminal==true'. This class and its subclasses are used to represent grammars that are terminals. */
  public static class Terminal extends Grammar {
    /** Constructs a Terminal which matches a single Token of type 'type'. */
    public Terminal(int type) { super(true); this.type = type; }
    
    /** The type of Tokens matched by this Terminal (one of the 'Token.TYPE_XXX' values). */
    public final int type;
    
    /** Returns 'true' if 't' is the right kind of Token for this Terminal. */
    public boolean accepts(Token t) { return this.type==t.type; }
    
    /** Returns a String representation of this Terminal. */
    public String toString() {
      switch (this.type) {
        case Token.TYPE_COMMENT: return "COMMENT";
        case Token.TYPE_WORD: return "WORD";
        case Token.TYPE_CONSTANT: return "CONSTANT";
        case Token.TYPE_IDENTIFIER: return "IDENTIFIER";
        case Token.TYPE_STRING: return "STRING";
        case Token.TYPE_NUMBER: return "NUMBER";
        case Token.TYPE_CHAR: return "CHAR";
        case Token.TYPE_ROUND: return "ROUND";
        case Token.TYPE_SQUARE: return "SQUARE";
        case Token.TYPE_BRACE: return "BRACE";
        default: throw new RuntimeException("Unknown Token type: "+this.type);
      }
    }

    /** If this Terminal is of type 'TYPE_BRACKET', 'TYPE_SQUARE' or 'TYPE_BRACE', override this method so that it returns the Grammar that should be used to parse the contents of any brackets that are accepted by this Terminal. Ideally, the same Grammar object should be returned every time the method is called. Do not call this method for Terminals that do not accept brackets. The default implementation throws a RuntimeException. Note that this method has no effect on the 'accepts()' method. */
    public Grammar getGrammarOfContents() {
      throw new RuntimeException("This Terminal does not accept brackets");
    }
  }
  
  ////////////////////////////////////////////////////////////////////////////
  
  /** A subclass of Terminal which insists on an exact text match with a Token of type 'Token.TYPE_WORD'. This is used for keywords, punctuation strings and separator characters. */
  public static class Keyword extends Terminal {
    /** Constructs a Keyword. */
    public Keyword(String text) { super(Token.TYPE_WORD); this.text = text; }
    
    /** The exact text to match. */
    public final String text;

    /** Returns 'true' if 't' is of type 'Token.TYPE_WORD' and its text is an exact match for 'text'. */
    public boolean accepts(Token t) {
      if (!super.accepts(t)) return false;
      if (this.text.length()!=t.length) return false;
      for (int i=0; i<t.length; i++) {
        if (this.text.charAt(i)!=t.sentence.get(t.start+i)) return false;
      }
      return true;
    }
    
    /** Returns 'text' enclosed in double quotes. */
    public String toString() { return '"'+this.text+'"'; }
  }
  
  ////////////////////////////////////////////////////////////////////////////
  
  /** The subclass of Grammar with 'isTerminal==false'. This class represents quantified non-terminals. By quantified I mean that it can be optional and/or repeatable (as represented in SSS grammar specifications by the symbols '?', '*' and '+'). This class is called 'NonTerminal' only as an abbreviation for 'QuantifiedNonTerminal'. */
  public static class NonTerminal extends Grammar {
    /** Constructs a NonTerminal.
     * @param prods an array of Productions that produce this NonTerminal, in order of preference. The array is not copied away, and must therefore not be modified hereafter.
     * @param isOptional 'true' if zero copies of this non-terminal is acceptable, or 'false' to insist on at least one copy.
     * @param isRepeatable 'true' if more than one copy of this non-terminal is acceptable, or 'false' to insist on at most one copy.
     */
    public NonTerminal(
      Production[] prods,
      boolean isOptional,
      boolean isRepeatable
    ) {
      super(false);
      this.prods = prods; this.numProds = prods.length;
      this.isOptional = isOptional; this.isRepeatable = isRepeatable;
    }
    
    /** Constructs a NonTerminal with the same productions as 'that' but different flags. */
    public NonTerminal(
      NonTerminal that,
      boolean isOptional,
      boolean isRepeatable
    ) {
      this(that.prods, isOptional, isRepeatable);
    }
    
    /** The number of Productions. */
    public final int numProds;
    
    /** Returns the Production with index 'index'. */
    public Production get(int index) { return this.prods[index]; }
    
    /** One of the flags passed to the constructor. */
    public final boolean isOptional, isRepeatable;
    
    /** Returns a String representation of this NonTerminal. */
    public String toString() {
      final StringBuffer ans = new StringBuffer("{");
      if (this.numProds>0) {
        ans.append(this.get(0));
        for (int i=1; i<this.numProds; i++)
          ans.append(" | ").append(this.get(i));
      }
      ans.append('}');
      if (this.isOptional) {
        if (this.isRepeatable)
          ans.append('*');
        else
          ans.append('?');
      } else {
        if (this.isRepeatable)
          ans.append('+');
      }
      return ans.toString();
    }
    
    private Production[] prods;
  }
  
  ////////////////////////////////////////////////////////////////////////////
  
  /** Represents a production of the grammar. A production gives a way of making a large syntactic thing out of smaller things. Every Production is identified by a 'name'. */
  public static class Production {
    /** Constructs a Production.
     * @param name the name of this Production.
     * @param parts an array of Grammars describing the things that are needed in order to use this Production.
     * @throws IllegalArgumentException if all of the parts are optional NonTerminals, because SSS forbids a production from matching the empty string. In particular, there must be at least one part.
     */
    public Production(String name, Grammar[] parts) {
      this.name = name;
      this.parts = parts; this.numParts = parts.length;
      boolean matchesEmpty = true;
      for (int i=0; matchesEmpty && i<parts.length; i++) {
        matchesEmpty &=
          !parts[i].isTerminal &&
          ((NonTerminal)parts[i]).isOptional;
      }
      if (matchesEmpty) throw new IllegalArgumentException(
        "Productions must not match the empty string"
      );
    }
    
    /** The name of this Production. Conventionally, this takes the form of an SSS constant, because it refers to something defined in a grammar specification. */
    public final String name;
    
    /** The number of parts. */
    public final int numParts;
    
    /** Returns the part with index 'index'. */
    public Grammar get(int index) { return this.parts[index]; }
    
    /** Returns a String representation of this Production. */
    public String toString() {
      final StringBuffer ans = new StringBuffer();
      ans.append(this.get(0));
      for (int i=1; i<this.numParts; i++) ans.append(' ').append(this.get(i));
      return ans.toString();
    }
    
    private Grammar[] parts;
  }
  
  ////////////////////////////////////////////////////////////////////////////
}
