package org.sc3d.apt.sss.v3;

import java.io.*;
import java.util.*;

/** A subclass of Parser which parses SSS grammar specifications. */
public class GrammarParser extends Parser {
  /** One GrammarParser will do for everybody, so this constructor is private. Use the 'PARSER' field instead. */
  private GrammarParser() { super(DECLARATION_PLUS); }
  
  /* New API. */
  
  /** A Grammar which represents what the specification of SSS grammars calls a 'grammar'.
   * <p>Note that this is not the Grammar of a grammar specification, which is what the specification of SSS grammars calls a 'declaration+'. */
  public static final Grammar.NonTerminal GRAMMAR =
    new Grammar.NonTerminal(new Grammar.Production[] {
      new Grammar.Production("GramComment", new Grammar[] {
        new Grammar.Terminal(Token.TYPE_COMMENT)
      }),
      new Grammar.Production("Comment", new Grammar[] {
        new Grammar.Keyword("COMMENT")
      }),
      new Grammar.Production("Word", new Grammar[] {
        new Grammar.Terminal(Token.TYPE_STRING)
      }),
      new Grammar.Production("Constant", new Grammar[] {
        new Grammar.Keyword("CONSTANT")
      }),
      new Grammar.Production("Identifier", new Grammar[] {
        new Grammar.Keyword("IDENTIFIER")
      }),
      new Grammar.Production("Round", new Grammar[] {
        new Grammar.Keyword("ROUND"),
        new GrammarInBracket(Token.TYPE_ROUND)
      }),
      new Grammar.Production("Square", new Grammar[] {
        new Grammar.Keyword("SQUARE"),
        new GrammarInBracket(Token.TYPE_ROUND)
      }),
      new Grammar.Production("Brace", new Grammar[] {
        new Grammar.Keyword("BRACE"),
        new GrammarInBracket(Token.TYPE_ROUND)
      }),
      new Grammar.Production("String", new Grammar[] {
        new Grammar.Keyword("STRING")
      }),
      new Grammar.Production("Number", new Grammar[] {
        new Grammar.Keyword("NUMBER")
      }),
      new Grammar.Production("Char", new Grammar[] {
        new Grammar.Keyword("CHAR")
      }),
      new Grammar.Production("Exact", new Grammar[] {
        new Grammar.Terminal(Token.TYPE_IDENTIFIER)
      }),
      new Grammar.Production("Option", new Grammar[] {
        new Grammar.Terminal(Token.TYPE_IDENTIFIER),
        new Grammar.Keyword("?")
      }),
      new Grammar.Production("Star", new Grammar[] {
        new Grammar.Terminal(Token.TYPE_IDENTIFIER),
        new Grammar.Keyword("*")
      }),
      new Grammar.Production("Plus", new Grammar[] {
        new Grammar.Terminal(Token.TYPE_IDENTIFIER),
        new Grammar.Keyword("+")
      })
    }, false, false);
  
  /** A Grammar which represents what the specification of SSS grammars calls a 'grammar+'. */
  public static final Grammar.NonTerminal GRAMMAR_PLUS =
    new Grammar.NonTerminal(GRAMMAR, false, true);
  
  /** A Grammar which represents what the specification of SSS grammars calls a 'production'. */
  public static final Grammar.NonTerminal PRODUCTION =
    new Grammar.NonTerminal(new Grammar.Production[] {
      new Grammar.Production("ProdComment", new Grammar[] {
        new Grammar.Terminal(Token.TYPE_COMMENT)
      }),
      new Grammar.Production("Production", new Grammar[] {
        new Grammar.Terminal(Token.TYPE_CONSTANT),
        new Grammar.Terminal(Token.TYPE_BRACE) {
          public Grammar getGrammarOfContents() { return GRAMMAR_PLUS; }
        }
      }),
    }, false, false);
  
  /** A Grammar which represents what the specification of SSS grammars calls a 'production+'. */
  public static final Grammar.NonTerminal PRODUCTION_PLUS =
    new Grammar.NonTerminal(PRODUCTION, false, true);
  
  /** A Grammar which represents what the specification of SSS grammars calls a 'declaration'. */
  public static final Grammar.NonTerminal DECLARATION =
    new Grammar.NonTerminal(new Grammar.Production[] {
      new Grammar.Production("DecComment", new Grammar[] {
        new Grammar.Terminal(Token.TYPE_COMMENT)
      }),
      new Grammar.Production("NonTerminal", new Grammar[] {
        new Grammar.Terminal(Token.TYPE_IDENTIFIER),
        new Grammar.Keyword("::="),
        new Grammar.Terminal(Token.TYPE_BRACE) {
          public Grammar getGrammarOfContents() { return PRODUCTION_PLUS; }
        }
      }),
      new Grammar.Production("Root", new Grammar[] {
        new Grammar.Keyword("ROOT"),
        GRAMMAR
      })
    }, false, false);
  
  /** A Grammar which represents what the specification of SSS grammars calls a 'declaration+'. */
  public static final Grammar.NonTerminal DECLARATION_PLUS =
    new Grammar.NonTerminal(DECLARATION, false, true);
  
  /** A GrammarParser. Use the value of this field instead of constructing a new GrammarParser. */
  public static final GrammarParser PARSER = new GrammarParser();
  
  /** Parses 'grammar' and returns a Grammar, or dies trying. If 'grammar' is not a syntactically correct grammar specification, this method prints diagnostic information to 'System.err' and throws an IllegalArgumentException. */
  public static Grammar fromString(String grammar) {
    final Sentence sentence = new Sentence(grammar.toCharArray());
    final Grammar ans = (Grammar)PARSER.parse(sentence);
    if (ans==null) {
      sentence.printErrorReport(System.err, 100);
      throw new IllegalArgumentException("Could not understand the grammar");
    }
    return ans;
  }
  
  /** Reads the named file, parses it, and returns a Grammar, or dies trying. If the file could not be read, or if it was not a syntactically correct grammar specification, this method prints diagnostic information to 'System.err' and throws an IllegalArgumentException. */
  public static Grammar fromFile(String filename) {
    try {
      return fromInputStream(new FileInputStream(filename));
    } catch (IOException e) {
      e.printStackTrace();
      throw new IllegalArgumentException("Could not open the grammar file");
    }
  }
  
  /** Reads a string of bytes from 'in', converts it to characters, parses it, and returns a Grammar, or dies trying. If anything goes wrong, this method prints diagnostic information to 'System.err' and throws an IllegalArgumentException. */
  public static Grammar fromInputStream(InputStream in) {
    try {
      final BufferedReader br = new BufferedReader(new InputStreamReader(in));
      final CharArrayWriter caw = new CharArrayWriter();
      final char[] cs = new char[100];
      while (true) {
        int used = br.read(cs);
        if (used==-1) break;
        caw.write(cs, 0, used);
      }
      br.close(); caw.close();
      final Sentence sentence = new Sentence(caw.toCharArray());
      final Grammar ans = (Grammar)PARSER.parse(sentence);
      if (ans==null) {
        sentence.printErrorReport(System.err, 100);
        throw new IllegalArgumentException("Could not understand the grammar");
      }
      return ans;
    } catch (IOException e) {
      e.printStackTrace();
      throw new IllegalArgumentException("Could not read the grammar");
    }
  }
  
  /* Override things in Parser. */
  
  /** Parses the contents of all brackets in the grammar file, and returns the Grammar specified by the 'ROOT' declaration.
   * @param raw a Tree representing a sentence of what the specification of SSS grammars calls a 'declaration+'.
   * @return a Grammar.
   */
  public Object postProcess(Tree raw) {
    final Tree.NonTerminal decs = (Tree.NonTerminal)raw;
    final Grammar.NonTerminal EMPTY =
      new Grammar.NonTerminal(new Grammar.Production[0], false, false);
    // Parse all of the NonTerminals in turn and find the Root.
    // Fill up a Map from identifiers, and a list of UpdatableBrackets.
    boolean isOkay = true;
    Tree.Production root = null;
    final HashMap defs = new HashMap();
    final LinkedList bs = new LinkedList();
    for (int i=0; i<decs.length; i++) {
      final Tree.Production p = decs.get(i);
      if (p.name.equals("DecComment")) {
        // Do nothing.
      } else if (p.name.equals("NonTerminal")) {
        final Token identifier = p.getT(0).t;
        final String name = identifier.toString();
        if (defs.containsKey(name)) {
          identifier.addError("This non-terminal is already defined.");
          isOkay = false;
        }
        final Tree raw2 = p.getT(2).parse();
        if (raw2==null) isOkay = false; else {
          final Grammar.Production[] prods =
            this.processProductions(defs, bs, raw2);
          if (prods==null) {
            defs.put(name, EMPTY);
            isOkay = false;
          } else {
            defs.put(name, new Grammar.NonTerminal(prods, false, false));
          }
        }
      } else if (p.name.equals("Root")) {
        if (root!=null) {
          root.getT(0).t.addError(
            "There may only be one root declaration (this is the first)."
          );
          p.getT(0).t.addError(
            "There may only be one root declaration (this is not the first)."
          );
          isOkay = false;
        } else {
          root = p;
        }
      } else {
        throw new RuntimeException("What to do with "+p.name+"?");
      }
    }
    // Calculate the Grammar corresponding to the root.
    Grammar ans = null;
    if (root==null) {
      final Sentence sentence = decs.get(0).getT(0).t.sentence;
      sentence.addError(
        "A grammar specification must have a ROOT declaration",
        0, sentence.length
      );
      isOkay = false;
    } else {
      final Grammar[] gs = this.processGrammars(defs, bs, root.get(1));
      if (gs==null) isOkay = false; else {
        if (gs.length!=1) throw new RuntimeException("Eh?");
        ans = gs[0];
      }
    }
    // Resolve all the UpdatableBrackets.
    while (bs.size()!=0) {
      final UpdatableBracket u = (UpdatableBracket)bs.removeLast();
      final Tree raw2 = u.spec.parse();
      if (raw2==null) isOkay = false; else {
        final Grammar[] gs = this.processGrammars(defs, bs, raw2);
        if (gs==null) isOkay = false; else {
          if (gs.length!=1) throw new RuntimeException("Eh?");
          u.gOfC = gs[0];
        }
      }
    }
    // Return.
    return isOkay ? ans : null;
  }
  
  /* Private. */
  
  /** Turns the raw parse-tree of a 'production*' into an array of Grammar.Productions. Comments are stripped out. It is an error if two or more of the Productions have the same name.
   * @param defs a Map from identifiers to Grammar.NonTerminals. This is used to resolve the names of non-terminals. It is not modified.
   * @param bs a list to which to append UpdatableBrackets that need to be resolved later.
   * @param raw a non-null Tree representing a sentence of what the specification of SSS grammars calls a 'production*'.
   * @return a Grammar.Production[], or 'null' if there is an error.
   */
  private Grammar.Production[] processProductions(Map defs, List bs, Tree raw) {
    // Collect the Productions.
    final Tree.NonTerminal prods = (Tree.NonTerminal)raw;
    final HashSet names = new HashSet();
    final LinkedList l = new LinkedList();
    boolean isOkay = true;
    for (int i=0; i<prods.length; i++) {
      final Tree.Production prod = prods.get(i);
      if (prod.name.equals("ProdComment")) {
        // Do nothing.
      } else if (prod.name.equals("Production")) {
        final Token constant = prod.getT(0).t;
        final String name = constant.toString();
        if (!names.add(name)) {
          constant.addError(
            "The productions of a non-terminal must have different "+
            "identifying constants."
          );
          isOkay = false;
        }
        final Tree.Terminal brace = prod.getT(1);
        final Tree raw2 = brace.parse();
        if (raw2==null) isOkay = false; else {
          final Grammar[] gs = this.processGrammars(defs, bs, raw2);
          if (gs==null) isOkay = false; else {
            boolean allAcceptEmptyString = true;
            for (int j=0; allAcceptEmptyString&&j<gs.length; j++) {
              allAcceptEmptyString &= gs[j] instanceof Grammar.Terminal
                ? false
                : ((Grammar.NonTerminal)gs[j]).isOptional;
            }
            if (allAcceptEmptyString) {
              brace.t.addError(
                "At least one of these parts must be a terminal or a "+
                "non-optional non-terminal. You cannot write a production "+
                "that accepts the empty string."
              );
              isOkay = false;
            }
            l.add(new Grammar.Production(name, gs));
          }
        }
      } else {
        throw new RuntimeException("What to do with '"+prod.name+"'?");
      }
    }
    if (!isOkay) return null;
    return (Grammar.Production[])l.toArray(new Grammar.Production[l.size()]);
  }
  
  /** Turns the raw parse-tree of a 'grammar*' into an array of Grammars. Comments are stripped out.
   * @param defs a Map from identifiers to Grammar.NonTerminals. This is used to resolve the names of non-terminals. It is not modified.
   * @param bs a list to which to append UpdatableBrackets that need to be resolved later.
   * @param raw a non-null Tree representing a sentence of what the specification of SSS grammars calls a 'grammar*'.
   * @return a Grammar[], or 'null' if there is an error.
   */
  private Grammar[] processGrammars(Map defs, List bs, Tree raw) {
    final Tree.NonTerminal grams = (Tree.NonTerminal)raw;
    final LinkedList l = new LinkedList();
    boolean isOkay = true;
    for (int i=0; i<grams.length; i++) {
      final Tree.Production gram = grams.get(i);
      if (gram.name.equals("GramComment")) {
        // Do nothing.
      } else if (gram.name.equals("Comment")) {
        l.add(new Grammar.Terminal(Token.TYPE_COMMENT));
      } else if (gram.name.equals("Word")) {
        final Token t = gram.getT(0).t;
        final String s = ((SSSString)t).value;
        final Lex lex = new Lex(new Sentence(s.toCharArray()));
        if (lex.numTokens!=1 || lex.getToken(0).type!=Token.TYPE_WORD) {
          t.addError(
            "You cannot use this as a keyword, because it would not lex as "+
            "a single keyword, separator or punctuation word token."
          );
          isOkay = false;
        } else {
          l.add(new Grammar.Keyword(lex.getToken(0).toString()));
        }
      } else if (gram.name.equals("Constant")) {
        l.add(new Grammar.Terminal(Token.TYPE_CONSTANT));
      } else if (gram.name.equals("Identifier")) {
        l.add(new Grammar.Terminal(Token.TYPE_IDENTIFIER));
      } else if (
        gram.name.equals("Round") ||
        gram.name.equals("Square") ||
        gram.name.equals("Brace")
      ) {
        final UpdatableBracket u = new UpdatableBracket(
          gram.name.equals("Round") ? Token.TYPE_ROUND :
          gram.name.equals("Square") ? Token.TYPE_SQUARE : Token.TYPE_BRACE,
          gram.getT(1)
        );
        l.add(u);
        bs.add(u);
      } else if (gram.name.equals("String")) {
        l.add(new Grammar.Terminal(Token.TYPE_STRING));
      } else if (gram.name.equals("Number")) {
        l.add(new Grammar.Terminal(Token.TYPE_NUMBER));
      } else if (gram.name.equals("Char")) {
        l.add(new Grammar.Terminal(Token.TYPE_CHAR));
      } else if (
        gram.name.equals("Exact") ||
        gram.name.equals("Option") ||
        gram.name.equals("Star") ||
        gram.name.equals("Plus")
      ) {
        final Token identifier = ((Tree.Terminal)gram.get(0)).t;
        final Grammar.NonTerminal exact =
          (Grammar.NonTerminal)defs.get(identifier.toString());
        if (exact==null) {
          identifier.addError(
            "This non-terminal is not defined. Non-terminals must be defined "+
            "before they are used."
          );
          isOkay = false;
        } else if (gram.name.equals("Option")) {
          l.add(new Grammar.NonTerminal(exact, true, false));
        } else if (gram.name.equals("Star")) {
          l.add(new Grammar.NonTerminal(exact, true, true));
        } else if (gram.name.equals("Plus")) {
          l.add(new Grammar.NonTerminal(exact, false, true));
        } else {
          l.add(exact);
        }
      } else {
        throw new RuntimeException("What to do with '"+gram.name+"'?");
      }
    }
    return isOkay ? (Grammar[])l.toArray(new Grammar[l.size()]) : null;
  }
  
  ////////////////////////////////////////////////////////////////////////////
  
  /** A subclass of Grammar.Terminal whose 'getGrammarOfContents()' method returns 'GRAMMAR'. */
  private static class GrammarInBracket extends Grammar.Terminal {
    public GrammarInBracket(int type) { super(type); }
    public Grammar getGrammarOfContents() { return GRAMMAR; }
  }
  
  ////////////////////////////////////////////////////////////////////////////
  
  /** A subclass of Grammar.Terminal whose 'getGrammarOfContents()' method returns the value of a field that can be modified after construction time. This is convenient when constructing Grammars that recurse via brackets. The privacy of this subclass ensures that the outside world does not get the wrong idea. */
  private static class UpdatableBracket extends Grammar.Terminal {
    public UpdatableBracket(int type, Tree.Terminal spec) {
      super(type);
      this.spec = spec;
    }

    public Grammar getGrammarOfContents() { return this.gOfC; }

    /** The field to fill in later. */
    public Grammar gOfC = null;

    /** The fragment that needs to be parsed to work out the value of 'gOfC'. */
    public Tree.Terminal spec;
  }

  ////////////////////////////////////////////////////////////////////////////
  
  /* Test code. */
  
  /** Takes one argument: a filename. Parses the file as an SSS grammar specification and prints out the resulting Grammar. */
  public static void main(String[] args) throws IOException {
    if (args.length!=1) throw new IllegalArgumentException(
      "Syntax: java org.sc3d.apt.sss.v3.GrammarParser <grammar filename>"
    );
    final Grammar g = fromFile(args[0]);
    System.out.println("grammar = "+g);
  }
}
