package org.sc3d.apt.sss.v3;

import java.util.*;

/** Represents a non-determinstic finite automaton for parsing a particular grammar. Any Grammar can be converted into an NDFA using the 'make()' method. Unless your requirements are unusual, you probably don't need to use NDFAs directly, as Parser wraps it all up for you.
 * <h3>Ingredients</h3>
 * <p>An NDFA consists of:<ul>
 * <li>a set of <em>states</em>, one of which is the <em>initial state</em> and one the <em>final State</em>, and
 * <li>a set of <em>transitions</em> between the states, each representing a step in a depth-first construction of a parse-tree.
 * </ul>
 * <h3>Parsing using an NDFA</h3>
 * <p>Every sentence of the grammar corresponds to sequence of transitions that leads from the initial state to the final state. If the construction steps represented by the transitions are performed in order, the effect is to construct the parse-tree of the sentence. Alternatively, reading only the 'shift' steps (see below) lists the Tokens of the sentence in order.
 * <p>Thus to parse a sentence it suffices to find a suitable path from the initial to the final state. To resolve ambiguity in the grammar, the transitions from each state are ordered from most preferred to least preferred.
 * <h3>Depth-first construction of a Tree</h3>
 * <p>For the purposes of this class, a depth-first construction of a Tree is one in which every non-terminal is constructed before its child Trees, and in which children are constructed in left-to-right order. Thus at an intermediate stage of the construction there may be many Trees, and the right-hand Tree at each level may not yet have all of its children.
 * <p>Each step in a depth-first construction takes such a list of Trees and applies a simple modification to it. The effect of all the steps together is to turn an empty list into a list containing exactly one tree, which is the final parse-tree.
 * <p>In addition to no-op transitions, which do nothing, there are three kinds of Transition:<ul>
 * <li>A <em>shift</em> transition adds a Tree.Terminal to the list.
 * <li>An <em>open</em> transition adds an empty Tree.NonTerminal to the list. The NonTerminal is empty in the sense that its 'length' field is zero.
 * <li>A <em>reduce</em> transition takes the right-hand few Trees and wraps them up as a Tree.Production, which it then adds as a child to the preceding Tree.NonTerminal.
 * </ul>Those shift transitions which construct Tree.Brackets provide an NDFA for parsing the contents of the brackets.
 * <h3>Representation</h3>
 * <p>States are represented by instances of NDFA.State, and transitions are represented by instances of NDFA.Transition. Every State knows the Transitions that lead from it, and every Transition knows the State it leads to. It is therefore possible to follow paths through an NDFA forwards only. Every NDFA knows its initial and final States, but States do not know to which NDFA they belong. It is therefore possible for NDFAs to share States.
 * <p>The various different kinds of Transition are all represented by instances of Transition, and are distinguished according to which fields are set to non-null values. For example, a Transition is a no-op if all of its fields (apart from 'to') are 'null'.
 */
public class NDFA {
  /** Constructs an NDFA, given its initial and final States. See also 'make()# and 'makeProduction()'. */
  public NDFA(State initialState, State finalState) {
    this.initialState = initialState;
    this.finalState = finalState;
  }
  
  /* New API. */
  
  /** Constructs and returns an NDFA representing 'grammar'. If 'grammar' is a Grammar.Terminal, the NDFA will be just a single shift transition. If 'grammar' is a Grammar.NonTerminal, then the NDFA will consist of one open transition followed by the parallel composition of parts constructed using 'makeProduction()' joined up with some no-op transitions.
   * <p>The returned NDFA will have the property that any two different paths with the same start and end points both contain shift transitions. In particular, every cycle includes at least one shift transition.
   * <p>The initial and final States for the new NDFA are passed as arguments. All other States are freshly constructed. If you want the initial and/or final States freshly constructed, just pass 'new State()'. Outgoing transitions are added to the initial State and for repeatable NonTerminals also to the final State.
   * @param init the desired initial State.
   * @param fin the desired final State.
   * @param grammar the Grammar to represent.
   */
  public static NDFA make(
    final State init,
    final State fin,
    final Grammar grammar
  ) {
    if (grammar.isTerminal) {
      final Grammar.Terminal g = (Grammar.Terminal)grammar;
      init.add(new Transition(fin, g));
    } else {
      final Grammar.NonTerminal g = (Grammar.NonTerminal)grammar;
      final State open = g.isOptional && g.isRepeatable ? fin : new State();
      init.add(new Transition(open, g));
      for (int i=0; i<g.numProds; i++) makeProduction(open, fin, g.get(i));
      if (g.isOptional && !g.isRepeatable)
        open.add(new Transition(fin));
      if (!g.isOptional && g.isRepeatable)
        fin.add(new Transition(open));
    }
    return new NDFA(init, fin);
  }
  
  /** Constructs and returns an NDFA representing 'production'. The NDFA will consist of a chain of parts constructed using 'make()' followed by a reduce transition.
   * <p>The returned NDFA will has the property that any two different paths with the same start and end points both contain shift transitions. In particular, every cycle includes at least one shift transition.
   * <p>The initial and final States for the new NDFA are passed as arguments. All other States are freshly constructed. If you want the initial and/or final States freshly constructed, just pass 'new State()'. Outgoing transitions are added to the initial State but not to the final State.
   * @param init the desired initial State.
   * @param fin the desired final State.
   * @param production the Grammar.Production to represent.
   */
  public static NDFA makeProduction(
    final State init,
    final State fin,
    final Grammar.Production production
  ) {
    State last = init;
    for (int i=0; i<production.numParts; i++)
      last = make(last, new State(), production.get(i)).finalState;
    last.add(new Transition(fin, production));
    return new NDFA(init, fin);
  }
  
  /** The initial State of this NDFA. */
  public final State initialState;
  
  /** The final State of this NDFA. */
  public final State finalState;
  
  /** Returns a String representation of this NDFA, including all States reachable from 'this.initialState'. */
  public String toString() {
    StringBuffer ans = new StringBuffer("NDFA[");
    ans.append("\n  initialState = 0");
    final HashMap done = new HashMap();
    final LinkedList toDo = new LinkedList();
    int count = 0;
    done.put(this.initialState, new Integer(0));
    toDo.addLast(this.initialState);
    while (toDo.size()!=0) {
      final State q = (State)toDo.removeLast();
      if (q==this.finalState)
        ans.append("\n  finalState = ").append(done.get(this.finalState));
      final Transition[] ts = q.toArray();
      for (int i=0; i<ts.length; i++) {
        if (!done.containsKey(ts[i].to)) {
          done.put(ts[i].to, new Integer(done.size()));
          toDo.addLast(ts[i].to);
        }
        ans.append("\n  from ").append(done.get(q));
        ans.append(" to ").append(done.get(ts[i].to));
        ans.append(" label ").append(ts[i]);
        count++;
      }
    }
    ans.append("\n  (total of ").append(count).append(" transitions)");
    return ans.append("\n]").toString();
  }
  
  ////////////////////////////////////////////////////////////////////////////
  
  /** Represents a state of an NDFA. Instances are not immutable: outgoing Transitions can be added using the 'add()' method. */
  public static class State {
    /** Constructs a State with no outgoing Transitions. */
    public State() {
      this.transitions = new Transition[10];
      this.used = 0;
    }
    
    /** Returns the number of outgoing Transitions added to far. */
    public int count() { return this.used; }
    
    /** Returns the outgoing Transition with index 'index'. Transitions are numbered from '0' (most preferred) to 'count()-1' (least preferred). */
    public Transition get(int index) {
      if (index>=this.used) throw new ArrayIndexOutOfBoundsException();
      return this.transitions[index];
    }
    
    /** Returns all the Transitions from this State in order of preference in a fresh array. */
    public Transition[] toArray() {
      final Transition[] ans = new Transition[this.used];
      System.arraycopy(this.transitions, 0, ans, 0, this.used);
      return ans;
    }
    
    /** Adds an outgoing Transition to this State. */
    public void add(Transition t) {
      if (this.used>=this.transitions.length) {
        final Transition[] old = this.transitions;
        this.transitions = new Transition[old.length*2];
        System.arraycopy(old, 0, this.transitions, 0, this.used);
      }
      this.transitions[this.used++] = t;
    }
    
    /* Private. */
    
    private Transition[] transitions;
    private int used;
  }
  
  ////////////////////////////////////////////////////////////////////////////
  
  /** Represents a transition of an NDFA. Instances are immutable, except for the 'ndfa' field which should be filled in by the caller when it is first needed. */
  public static class Transition {
    /** Constructs a Transition, given values for its fields. */
    private Transition(State to, Grammar g, Grammar.Production p) {
      this.to = to; this.g = g; this.p = p;
    }

    /** Constructs a 'no-op' Transition, given values for its non-null fields. */
    public Transition(State to) { this(to, null, null); }
    
    /** Constructs a 'shift' Transition (bracket or otherwise), given values for its non-null fields. */
    public Transition(State to, Grammar.Terminal g) { this(to, g, null); }
    
    /** Constructs an 'open' Transition, given values for its non-null fields. */
    public Transition(State to, Grammar.NonTerminal g) { this(to, g, null); }
    
    /** Constructs a 'reduce' Transition, given values for its non-null fields. */
    public Transition(State to, Grammar.Production p) { this(to, null, p); }
    
    /** The State to which this Transition leads. This field is never null. */
    public final State to;
    
    /** A Grammar describing the Tree that is constructed by following this Transition, or 'null'. Specifically:<ul>
     * <li>If this is a 'shift' Transition, this field holds a Grammar.Terminal (perhaps a Grammar.Bracket).
     * <li>If this is an 'open' Transition, this field holds a Grammar.NonTerminal.
     * <li>If this is a 'no-op' or 'reduce' Transition, this field is null.
     * </ul> */
    public final Grammar g;
    
    /** A Grammar.Production to apply when following this 'reduce' Transition, or 'null' if this is not a 'reduce' Transition. */
    public final Grammar.Production p;
    
    /** An NDFA describing the grammar of the contents of the bracket, if this is a 'bracket shift' Transition and its NDFA has already been calculated, or 'null' if this is not a 'bracket shift' Transition or the value has not yet been calculated. Calculating the NDFA is the caller's responsibility, but it involves nothing more complicated than calling 'make()'. */
    public NDFA ndfa = null;
    
    /** Returns a brief decription of this Transition. Specifically, returns "No-op" or "Open", or "Shift[x]" where "x" is 'g.toString()', or "Reduce[x]" where "x" is 'p.name'. */
    public String toString() {
      if (this.g==null) {
        if (this.p==null) return "No-op";
        return "Reduce["+this.p.name+"]";
      } else {
        if (!this.g.isTerminal) return "Open";
        return "Shift["+this.g.toString()+"]";
      }
    }
  }
}
