package org.sc3d.apt.sss.v3;

/** A subclass of Token that represents a literal character. The 'type' field is always 'Token.TYPE_CHAR'. This class adds a 'value' field that gives the unicode character value of this literal character. */
public class SSSChar extends Token {
  /** Constructs an SSSChar given values for its fields. */
  public SSSChar(
    Sentence sentence, int start, int length, int value, boolean isOkay
  ) {
    super(Token.TYPE_CHAR, sentence, start, length);
    this.value = value; this.isOkay = isOkay;
  }
  
  /* New API. */
  
  /** The unicode character value of this literal character. This value is undefined unless 'isOkay'. */
  public final int value;
  
  /** 'true' if this literal character is syntactically correct, otherwise 'false'. */
  public final boolean isOkay;
  
  /** Lexes a single SSSChar from 'sentence' starting at 'start'. This method always returns an SSSChar. If there is a syntax error, this method returns an SSSChar representing the correct part, and attaches error messages to 'sentence'.
   * @param sentence the sentence from which to pluck an SSSChar.
   * @param start the index of the opening ' character.
   * @return a fresh SSSChar.
   * @throws IllegalArgumentException if there isn't a ' character at 'start'.
   */
  public static SSSChar lex(final Sentence sentence, final int start) {
    int i = start;
    if (i>=sentence.length || sentence.get(i++)!='\'') {
      throw new IllegalArgumentException(
        "Literal character must start with a ' character"
      );
    }
    boolean isOkay = true;
    int value = 0;
    if (i<sentence.length) {
      final char c = sentence.get(i);
      if (c=='\\') { // Escape sequence.
        final Escape e = Escape.lex(sentence, i);
        if (e.isOkay) value = e.value; else isOkay = false;
        i += e.length;
      } else { // Ordinary character.
        value = c;
        i++;
      }
    }
    if (i<sentence.length && sentence.get(i)=='\'') {
      i++;
    } else {
      sentence.addError(
        "Bad character literal: the closing ' character is missing.",
        start, i-start
      );
      isOkay = false;
    }
    return new SSSChar(sentence, start, i-start, value, isOkay);
  }
  
  /* Test code. */
  
  public static void main(String[] args) {
    final String[] tests = new String[] {
      "'", "'\\", "'\\A", "'\\A\\", "'\\A\\'", "'A", "'A'", "'AA", "'AA'"
    };
    for (int i=0; i<tests.length; i++) {
      System.out.println("\ntests["+i+"] = "+tests[i]);
      final Sentence sentence = new Sentence(tests[i].toCharArray());
      final SSSChar me = SSSChar.lex(sentence, 0);
      System.out.println("me = "+me);
      System.out.println("me.value = "+me.value);
      System.out.println("me.isOkay = "+me.isOkay);
      sentence.printErrorReport(System.out, 1000);
    }
  }
}
