package org.sc3d.apt.sss.v3;

import java.io.*;

/** Represents a sequence of characters that need to be parsed, and allows error messages to be attached to it. */
public class Sentence {
  /** Constructs a Sentence given an array of characters. */
  public Sentence(char[] cs) {
    this.cs = cs;
    this.length = cs.length;
    this.es = new Error[((this.length+1)>>ERROR_BLOCK)+1];
  }
  
  /* New API. */
  
  /** The number of characters in this Sentence. */
  public final int length;
  
  /** Returns the character at 'index'. */
  public char get(int index) { return this.cs[index]; }
  
  /** Returns part of this Sentence as a String.
   * @param index the index at which the desired String starts.
   * @param length the length of the desired String.
   */
  public String getString(int index, int length) {
    return new String(this.cs, index, length);
  }
  
  /** Attaches an error message to part of this Sentence. This method may do nothing if there is already an error message attached nearby.
   * @param message the error message.
   * @param index the index of the first character to which the error message applies. Must be in the range from '0' to 'this.length' inclusive.
   * @param length the number of characters to which this error message applies.
   */
  public void addError(String message, int index, int length) {
    final int n = index>>ERROR_BLOCK;
    if (this.es[n]!=null) return;
    this.es[n] = new Error(message, index, length);
    this.errorCount++;
  }
  
  /** Equivalent to 'addError(message, length, 0). */
  public void addErrorAtEnd(String message) {
    this.addError(message, this.length, 0);
  }
  
  /** Returns the number of errors attached to this Sentence. */
  public int countErrors() { return this.errorCount; }
  
  /** Prints an error report to 'out'.
   * @param out a PrintStream to which to print the error report.
   * @param maxErrors the maximum number of errors to print.
   */
  public void printErrorReport(PrintStream out, int maxErrors) {
    int line = 1, start = 0;
    for (int i=0; i<=this.length; i++) {
      if (i==this.length || this.get(i)=='\n') {
        for (int j=start>>ERROR_BLOCK; j<=i>>ERROR_BLOCK; j++) {
          final Error e = this.es[j];
          if (e!=null && e.index>=start && e.index<=i) {
            out.println("Line "+line+":");
            out.println(this.getString(start, i-start));
            for (int k=start; k<e.index; k++)
              out.print(this.get(k)=='\t' ? '\t' : ' ');
            final int end = e.index+e.length;
            for (int k=e.index; k<i && k<end; k++) out.print('^');
            if (e.length==0 || i+1==end) out.print('^');
            out.println();
            if (i+1<end)
              out.println("(continues for another "+(end-i)+" characters)");
            out.println(e.message);
            out.println();
            if (--maxErrors<=0) return;
          }
        }
        line++;
        start = i+1;
      }
    }
  }
  
  /* Private. */
  
  /** Represents an Error message. */
  private static class Error {
    Error(String message, int index, int length) {
      this.message = message; this.index = index; this.length = length;
    }
    final String message;
    final int index, length;
  }
  
  /** The characters of this Sentence. */
  private final char[] cs;
  
  /** The log to the base 2 of the number of characters in each error neighbourhood. This value dictates how close together errors can be. If two errors are attached in the same neighbourhood, only the earlier one is stored. */
  private static final int ERROR_BLOCK = 4;
  
  /** The error messages attached to this Sentence. An error message which starts at 'index' is stored in element 'index>>ERROR_BLOCK' of this array. */
  private final Error[] es;
  
  /** Records the number of non-null elements of 'es'. */
  private int errorCount;
  
  /* Test code. */
  
  /** Reads the named file and returns it as a Sentence. */
  public static Sentence readFile(String filename) throws IOException {
    final FileReader fr = new FileReader(filename);
    final CharArrayWriter caw = new CharArrayWriter();
    final char[] cs = new char[100];
    while (true) {
      int used = fr.read(cs);
      if (used==-1) break;
      caw.write(cs, 0, used);
    }
    return new Sentence(caw.toCharArray());
  }
  
  /** Takes one argument: a filename. Reads the file, adds errors at random, then prints an error report. */
  public static void main(String[] args) throws IOException {
    if (args.length!=1) throw new IllegalArgumentException(
      "Syntax: java org.sc3d.apt.sss.v3.Sentence <filename>"
    );
    final Sentence me = readFile(args[0]);
    final java.util.Random rand = new java.util.Random();
    for (int i=0; i<20; i++) {
      final int p = rand.nextInt(me.length), q = rand.nextInt(me.length);
      if (p<q) me.addError("Error message "+i, p, q-p);
    }
    me.printErrorReport(System.out, 1000000);
  }
}
