package org.sc3d.apt.sss.v3;

/** Represents a part of a Sentence that has some sort of syntactic significance. For example, Tokens are used to represent keywords, comments and numbers. Every Token has a 'type' field whose value indicates what kind of Token it is. */
public class Token {
  /** Constructs a Token.
   * @param type one of the 'TYPE_XXX' values, indicating what kind of Token this is.
   */
  public Token(int type, Sentence sentence, int start, int length) {
    this.type = type; this.sentence = sentence;
    this.start = start; this.length = length;
  }
  
  /* New API. */
  
  /** The type of this Token (one of the 'TYPE_XXX' values). */
  public final int type;
  
  /** The Sentence of which this Token is part. */
  public final Sentence sentence;
  
  /** The index into 'sentence' of the character at which this Token starts. */
  public final int start;
  
  /** The length of this Token in characters. */
  public final int length;
  
  /** Attaches an error message to this Token. This method calls 'sentence.addError()'. */
  public void addError(String message) {
    this.sentence.addError(message, this.start, this.length);
  }
  
  /* Statics. */
  
  /** The value of the 'type' field for a Token that represents an SSS comment. */
  public static final int TYPE_COMMENT = 1;

  /** The value of the 'type' field for a Token that represents an SSS keyword, separator character, or punctuation string. */
  public static final int TYPE_WORD = 2;
  
  /** The value of the 'type' field for a Token that represents an SSS constant. */
  public static final int TYPE_CONSTANT = 3;
  
  /** The value of the 'type' field for a Token that represents an SSS identifier. */
  public static final int TYPE_IDENTIFIER = 4;
  
  /** The value of the 'type' field for a Token that represents an SSS literal string. */
  public static final int TYPE_STRING = 5;
  
  /** The value of the 'type' field for a Token that represents an SSS literal number. */
  public static final int TYPE_NUMBER = 6;
  
  /** The value of the 'type' field for a Token that represents an SSS literal character. */
  public static final int TYPE_CHAR = 7;
  
  /** The value of the 'type' field for a Token that represents a string of Tokens inside round brackets. */
  public static final int TYPE_ROUND = 8;

  /** The value of the 'type' field for a Token that represents a string of Tokens inside square brackets. */
  public static final int TYPE_SQUARE = 9;

  /** The value of the 'type' field for a Token that represents a string of Tokens inside braces (curly brackets). */
  public static final int TYPE_BRACE = 10;

  /* Override things in Object. */
  
  /** Returns the characters that make up this Token. */
  public final String toString() {
    return this.sentence.getString(this.start, this.length);
  }
  
  /** Returns 'this.toString().hashCode()'. */
  public int hashCode() { return this.toString().hashCode(); }
  
  /** Returns 'true' only if 'that' is a non-null Token whose 'toString()' method returns a String equal to 'this.toString()'. */
  public boolean equals(Object that) {
    return
      that!=null &&
      that instanceof Token &&
      this.toString().equals(that.toString());
  }
}
